# This is DFace's implementation of MTCNN modified for AutoKeras
# Link to DFace: https://github.com/kuaikuaikim/DFace
import os

import cv2
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torchvision.transforms as transforms
from torch.autograd.variable import Variable

from autokeras.constant import Constant
from autokeras.pretrained.base import Pretrained
from autokeras.utils import get_device, download_file_from_google_drive, temp_path_generator, ensure_dir


def weights_init(m):
    if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight.data)
        nn.init.constant_(m.bias, 0.1)


class PNet(nn.Module):

    def __init__(self):
        super(PNet, self).__init__()

        self.pre_layer = nn.Sequential(
            nn.Conv2d(3, 10, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(10, 16, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.Conv2d(16, 32, kernel_size=3, stride=1),
            nn.PReLU()
        )
        self.conv4_1 = nn.Conv2d(32, 1, kernel_size=1, stride=1)
        self.conv4_2 = nn.Conv2d(32, 4, kernel_size=1, stride=1)
        self.conv4_3 = nn.Conv2d(32, 10, kernel_size=1, stride=1)

        self.apply(weights_init)

    def forward(self, x):
        x = self.pre_layer(x)
        label = torch.sigmoid(self.conv4_1(x))
        offset = self.conv4_2(x)
        return label, offset


class RNet(nn.Module):

    def __init__(self):
        super(RNet, self).__init__()

        self.pre_layer = nn.Sequential(
            nn.Conv2d(3, 28, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(28, 48, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(48, 64, kernel_size=2, stride=1),
            nn.PReLU()

        )
        self.conv4 = nn.Linear(64 * 2 * 2, 128)
        self.prelu4 = nn.PReLU()
        self.conv5_1 = nn.Linear(128, 1)
        self.conv5_2 = nn.Linear(128, 4)
        self.conv5_3 = nn.Linear(128, 10)
        self.apply(weights_init)

    def forward(self, x):
        x = self.pre_layer(x)
        x = x.view(x.size(0), -1)
        x = self.conv4(x)
        x = self.prelu4(x)
        det = torch.sigmoid(self.conv5_1(x))
        box = self.conv5_2(x)
        return det, box


class ONet(nn.Module):

    def __init__(self):
        super(ONet, self).__init__()

        self.pre_layer = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(32, 64, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 64, kernel_size=3, stride=1),
            nn.PReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(64, 128, kernel_size=2, stride=1),
            nn.PReLU()
        )
        self.conv5 = nn.Linear(128 * 2 * 2, 256)
        self.prelu5 = nn.PReLU()
        self.conv6_1 = nn.Linear(256, 1)
        self.conv6_2 = nn.Linear(256, 4)
        self.conv6_3 = nn.Linear(256, 10)
        self.apply(weights_init)

    def forward(self, x):
        x = self.pre_layer(x)
        x = x.view(x.size(0), -1)
        x = self.conv5(x)
        x = self.prelu5(x)
        det = torch.sigmoid(self.conv6_1(x))
        box = self.conv6_2(x)
        landmark = self.conv6_3(x)
        return det, box, landmark


def get_square_bbox(bbox):
    square_bbox = bbox.copy()

    h = bbox[:, 3] - bbox[:, 1] + 1
    w = bbox[:, 2] - bbox[:, 0] + 1
    l = np.maximum(h, w)
    square_bbox[:, 0] = bbox[:, 0] + w * 0.5 - l * 0.5
    square_bbox[:, 1] = bbox[:, 1] + h * 0.5 - l * 0.5

    square_bbox[:, 2] = square_bbox[:, 0] + l - 1
    square_bbox[:, 3] = square_bbox[:, 1] + l - 1
    return square_bbox


def generate_bounding_box(map, reg, scale, threshold):
    stride = 2
    cellsize = 12

    t_index = np.where(map > threshold)

    if t_index[0].size == 0:
        return np.array([])

    dx1, dy1, dx2, dy2 = [reg[0, t_index[0], t_index[1], i] for i in range(4)]
    reg = np.array([dx1, dy1, dx2, dy2])

    score = map[t_index[0], t_index[1], 0]
    boundingbox = np.vstack([np.round((stride * t_index[1]) / scale),
                             np.round((stride * t_index[0]) / scale),
                             np.round((stride * t_index[1] + cellsize) / scale),
                             np.round((stride * t_index[0] + cellsize) / scale),
                             score,
                             reg
                             ])

    return boundingbox.T


def resize_image(img, scale):
    height, width, channels = img.shape
    new_height = int(height * scale)
    new_width = int(width * scale)
    new_dim = (new_width, new_height)
    img_resized = cv2.resize(img, new_dim, interpolation=cv2.INTER_LINEAR)
    return img_resized


def pad(bboxes, w, h):
    tmpw = (bboxes[:, 2] - bboxes[:, 0] + 1).astype(np.int32)
    tmph = (bboxes[:, 3] - bboxes[:, 1] + 1).astype(np.int32)
    numbox = bboxes.shape[0]

    dx = np.zeros((numbox,))
    dy = np.zeros((numbox,))
    edx, edy = tmpw.copy() - 1, tmph.copy() - 1

    x, y, ex, ey = bboxes[:, 0], bboxes[:, 1], bboxes[:, 2], bboxes[:, 3]

    tmp_index = np.where(ex > w - 1)
    edx[tmp_index] = tmpw[tmp_index] + w - 2 - ex[tmp_index]
    ex[tmp_index] = w - 1

    tmp_index = np.where(ey > h - 1)
    edy[tmp_index] = tmph[tmp_index] + h - 2 - ey[tmp_index]
    ey[tmp_index] = h - 1

    tmp_index = np.where(x < 0)
    dx[tmp_index] = 0 - x[tmp_index]
    x[tmp_index] = 0

    tmp_index = np.where(y < 0)
    dy[tmp_index] = 0 - y[tmp_index]
    y[tmp_index] = 0

    return_list = [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph]
    return_list = [item.astype(np.int32) for item in return_list]

    return return_list


def nms(dets, thresh, mode="Union"):
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        if mode == "Union":
            ovr = inter / (areas[i] + areas[order[1:]] - inter)
        elif mode == "Minimum":
            ovr = inter / np.minimum(areas[i], areas[order[1:]])

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep


def convert_image_to_tensor(image):
    transform = transforms.ToTensor()
    return transform(image)


def convert_chw_tensor_to_hwc_numpy(tensor):
    if isinstance(tensor, Variable):
        return np.transpose(tensor.data.numpy(), (0, 2, 3, 1))
    elif isinstance(tensor, torch.FloatTensor):
        return np.transpose(tensor.numpy(), (0, 2, 3, 1))
    else:
        raise Exception("covert b*c*h*w tensor to b*h*w*c numpy error.This tensor must have 4 dimension.")


def vis_face(im_array, dets, output_file_path, landmarks=None):
    fig, ax = plt.subplots(1)
    ax.imshow(im_array)

    for i in range(dets.shape[0]):
        bbox = dets[i, :4]

        rect = plt.Rectangle((bbox[0], bbox[1]),
                             bbox[2] - bbox[0],
                             bbox[3] - bbox[1], fill=False,
                             edgecolor='yellow', linewidth=0.9)
        ax.add_patch(rect)

    if landmarks is not None:
        for i in range(landmarks.shape[0]):
            landmarks_one = landmarks[i, :]
            landmarks_one = landmarks_one.reshape((5, 2))
            for j in range(5):
                cir1 = patches.Circle(xy=(landmarks_one[j, 0], landmarks_one[j, 1]), radius=2, alpha=0.4, color="red")
                ax.add_patch(cir1)
        plt.axis('off')
        fig.savefig(output_file_path, bbox_inches='tight', pad_inches=0)


class FaceDetector(Pretrained):
    """A class to predict faces using the MTCNN pre-trained model.
    """

    def __init__(self):
        super(FaceDetector, self).__init__()
        pnet, rnet, onet = self.load()
        self.device = get_device()

        self.pnet_detector = PNet()
        if torch.cuda.is_available():
            self.pnet_detector.load_state_dict(torch.load(pnet))
        else:
            self.pnet_detector.load_state_dict(torch.load(pnet, map_location=lambda storage, loc: storage))
        self.pnet_detector = self.pnet_detector.to(self.device)
        self.pnet_detector.eval()

        self.rnet_detector = RNet()
        if torch.cuda.is_available():
            self.rnet_detector.load_state_dict(torch.load(rnet))
        else:
            self.rnet_detector.load_state_dict(torch.load(rnet, map_location=lambda storage, loc: storage))
        self.rnet_detector = self.rnet_detector.to(self.device)
        self.rnet_detector.eval()

        self.onet_detector = ONet()
        if torch.cuda.is_available():
            self.onet_detector.load_state_dict(torch.load(onet))
        else:
            self.onet_detector.load_state_dict(torch.load(onet, map_location=lambda storage, loc: storage))
        self.onet_detector = self.onet_detector.to(self.device)
        self.onet_detector.eval()

        self.min_face_size = 24
        self.stride = 2
        self.threshold = [0.6, 0.7, 0.7]
        self.scale_factor = 0.709

    def load(self, model_path=None):
        temp_path = temp_path_generator()
        ensure_dir(temp_path)
        model_paths = [f'{temp_path}/{file_name}' for file_name in Constant.FACE_DETECTOR['MODEL_NAMES']]
        for google_id, file_name in zip(Constant.FACE_DETECTOR['MODEL_GOOGLE_ID'],
                                        Constant.FACE_DETECTOR['MODEL_NAMES']):
            download_file_from_google_drive(file_id=google_id, dest_path=f'{temp_path}/{file_name}')
        return model_paths

    def predict(self, img_path, output_file_path=None):
        """Predicts faces in an image.

        Args:
            img_path: A string. The path to the image on which the prediction is to be done.
            output_file_path: A string. The path where the output image is to be saved after the prediction. `None` by default.

        Returns:
            A tuple containing numpy arrays of bounding boxes and landmarks. Bounding boxes are of shape `(n, 5)` and
            landmarks are of shape `(n, 10)` where `n` is the number of faces predicted. Each bounding box is of length
            5 and the corresponding rectangle is defined by the first four values. Each bounding box has five landmarks
            represented by 10 coordinates.
        """
        if not os.path.exists(img_path):
            raise ValueError('Image does not exist')
        img = cv2.imread(img_path)
        img_bg = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        bounding_boxes, landmarks = self.detect_face(img)
        if output_file_path is not None:
            vis_face(img_bg, bounding_boxes, output_file_path, landmarks)
        return bounding_boxes, landmarks

    def detect_pnet(self, im):
        h, w, c = im.shape
        net_size = 12

        current_scale = float(net_size) / self.min_face_size
        im_resized = resize_image(im, current_scale)
        current_height, current_width, _ = im_resized.shape

        all_boxes = list()
        while min(current_height, current_width) > net_size:
            feed_imgs = []
            image_tensor = convert_image_to_tensor(im_resized)
            feed_imgs.append(image_tensor)
            feed_imgs = torch.stack(feed_imgs)
            feed_imgs = Variable(feed_imgs)

            if torch.cuda.is_available():
                feed_imgs = feed_imgs.cuda()

            cls_map, reg = self.pnet_detector(feed_imgs)

            cls_map_np = convert_chw_tensor_to_hwc_numpy(cls_map.cpu())
            reg_np = convert_chw_tensor_to_hwc_numpy(reg.cpu())

            boxes = generate_bounding_box(cls_map_np[0, :, :], reg_np, current_scale, self.threshold[0])

            current_scale *= self.scale_factor
            im_resized = resize_image(im, current_scale)
            current_height, current_width, _ = im_resized.shape

            if boxes.size == 0:
                continue
            keep = nms(boxes[:, :5], 0.5, 'Union')
            boxes = boxes[keep]
            all_boxes.append(boxes)

        if len(all_boxes) == 0:
            return None, None

        all_boxes = np.vstack(all_boxes)

        keep = nms(all_boxes[:, 0:5], 0.7, 'Union')
        all_boxes = all_boxes[keep]

        bw = all_boxes[:, 2] - all_boxes[:, 0] + 1
        bh = all_boxes[:, 3] - all_boxes[:, 1] + 1

        boxes = np.vstack([all_boxes[:, 0],
                           all_boxes[:, 1],
                           all_boxes[:, 2],
                           all_boxes[:, 3],
                           all_boxes[:, 4]
                           ])

        boxes = boxes.T

        align_topx = all_boxes[:, 0] + all_boxes[:, 5] * bw
        align_topy = all_boxes[:, 1] + all_boxes[:, 6] * bh
        align_bottomx = all_boxes[:, 2] + all_boxes[:, 7] * bw
        align_bottomy = all_boxes[:, 3] + all_boxes[:, 8] * bh

        boxes_align = np.vstack([align_topx,
                                 align_topy,
                                 align_bottomx,
                                 align_bottomy,
                                 all_boxes[:, 4]
                                 ])
        boxes_align = boxes_align.T

        return boxes, boxes_align

    def detect_rnet(self, im, dets):
        h, w, c = im.shape

        if dets is None:
            return None, None

        dets = get_square_bbox(dets)
        dets[:, 0:4] = np.round(dets[:, 0:4])

        [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph] = pad(dets, w, h)
        num_boxes = dets.shape[0]

        cropped_ims_tensors = []
        for i in range(num_boxes):
            tmp = np.zeros((tmph[i], tmpw[i], 3), dtype=np.uint8)
            tmp[dy[i]:edy[i] + 1, dx[i]:edx[i] + 1, :] = im[y[i]:ey[i] + 1, x[i]:ex[i] + 1, :]
            crop_im = cv2.resize(tmp, (24, 24))
            crop_im_tensor = convert_image_to_tensor(crop_im)
            cropped_ims_tensors.append(crop_im_tensor)
        feed_imgs = Variable(torch.stack(cropped_ims_tensors))

        if torch.cuda.is_available():
            feed_imgs = feed_imgs.to(self.device)

        cls_map, reg = self.rnet_detector(feed_imgs)

        cls_map = cls_map.cpu().data.numpy()
        reg = reg.cpu().data.numpy()

        keep_inds = np.where(cls_map > self.threshold[1])[0]

        if len(keep_inds) > 0:
            boxes = dets[keep_inds]
            cls = cls_map[keep_inds]
            reg = reg[keep_inds]
        else:
            return None, None

        keep = nms(boxes, 0.7)

        if len(keep) == 0:
            return None, None

        keep_cls = cls[keep]
        keep_boxes = boxes[keep]
        keep_reg = reg[keep]

        bw = keep_boxes[:, 2] - keep_boxes[:, 0] + 1
        bh = keep_boxes[:, 3] - keep_boxes[:, 1] + 1

        boxes = np.vstack([keep_boxes[:, 0],
                           keep_boxes[:, 1],
                           keep_boxes[:, 2],
                           keep_boxes[:, 3],
                           keep_cls[:, 0]
                           ])

        align_topx = keep_boxes[:, 0] + keep_reg[:, 0] * bw
        align_topy = keep_boxes[:, 1] + keep_reg[:, 1] * bh
        align_bottomx = keep_boxes[:, 2] + keep_reg[:, 2] * bw
        align_bottomy = keep_boxes[:, 3] + keep_reg[:, 3] * bh

        boxes_align = np.vstack([align_topx,
                                 align_topy,
                                 align_bottomx,
                                 align_bottomy,
                                 keep_cls[:, 0]
                                 ])

        boxes = boxes.T
        boxes_align = boxes_align.T

        return boxes, boxes_align

    def detect_onet(self, im, dets):
        h, w, c = im.shape

        if dets is None:
            return None, None

        dets = get_square_bbox(dets)
        dets[:, 0:4] = np.round(dets[:, 0:4])

        [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph] = pad(dets, w, h)
        num_boxes = dets.shape[0]

        cropped_ims_tensors = []
        for i in range(num_boxes):
            tmp = np.zeros((tmph[i], tmpw[i], 3), dtype=np.uint8)
            tmp[dy[i]:edy[i] + 1, dx[i]:edx[i] + 1, :] = im[y[i]:ey[i] + 1, x[i]:ex[i] + 1, :]
            crop_im = cv2.resize(tmp, (48, 48))
            crop_im_tensor = convert_image_to_tensor(crop_im)
            cropped_ims_tensors.append(crop_im_tensor)
        feed_imgs = Variable(torch.stack(cropped_ims_tensors))

        if torch.cuda.is_available():
            feed_imgs = feed_imgs.to(self.device)

        cls_map, reg, landmark = self.onet_detector(feed_imgs)

        cls_map = cls_map.cpu().data.numpy()
        reg = reg.cpu().data.numpy()
        landmark = landmark.cpu().data.numpy()

        keep_inds = np.where(cls_map > self.threshold[2])[0]

        if len(keep_inds) > 0:
            boxes = dets[keep_inds]
            cls = cls_map[keep_inds]
            reg = reg[keep_inds]
            landmark = landmark[keep_inds]
        else:
            return None, None

        keep = nms(boxes, 0.7, mode="Minimum")

        if len(keep) == 0:
            return None, None

        keep_cls = cls[keep]
        keep_boxes = boxes[keep]
        keep_reg = reg[keep]
        keep_landmark = landmark[keep]

        bw = keep_boxes[:, 2] - keep_boxes[:, 0] + 1
        bh = keep_boxes[:, 3] - keep_boxes[:, 1] + 1

        align_topx = keep_boxes[:, 0] + keep_reg[:, 0] * bw
        align_topy = keep_boxes[:, 1] + keep_reg[:, 1] * bh
        align_bottomx = keep_boxes[:, 2] + keep_reg[:, 2] * bw
        align_bottomy = keep_boxes[:, 3] + keep_reg[:, 3] * bh

        align_landmark_topx = keep_boxes[:, 0]
        align_landmark_topy = keep_boxes[:, 1]

        boxes_align = np.vstack([align_topx,
                                 align_topy,
                                 align_bottomx,
                                 align_bottomy,
                                 keep_cls[:, 0]
                                 ])

        boxes_align = boxes_align.T

        landmark = np.vstack([
            align_landmark_topx + keep_landmark[:, 0] * bw,
            align_landmark_topy + keep_landmark[:, 1] * bh,
            align_landmark_topx + keep_landmark[:, 2] * bw,
            align_landmark_topy + keep_landmark[:, 3] * bh,
            align_landmark_topx + keep_landmark[:, 4] * bw,
            align_landmark_topy + keep_landmark[:, 5] * bh,
            align_landmark_topx + keep_landmark[:, 6] * bw,
            align_landmark_topy + keep_landmark[:, 7] * bh,
            align_landmark_topx + keep_landmark[:, 8] * bw,
            align_landmark_topy + keep_landmark[:, 9] * bh,
        ])

        landmark_align = landmark.T

        return boxes_align, landmark_align

    def detect_face(self, img):
        boxes_align = np.array([])
        landmark_align = np.array([])

        if self.pnet_detector:
            boxes, boxes_align = self.detect_pnet(img)
            if boxes_align is None:
                return np.array([]), np.array([])

        if self.rnet_detector:
            boxes, boxes_align = self.detect_rnet(img, boxes_align)
            if boxes_align is None:
                return np.array([]), np.array([])

        if self.onet_detector:
            boxes_align, landmark_align = self.detect_onet(img, boxes_align)
            if boxes_align is None:
                return np.array([]), np.array([])

        return boxes_align, landmark_align



from autokeras.image.image_supervised import ImageClassifier, ImageRegressor, PortableImageSupervised
from autokeras.text.text_supervised import TextClassifier, TextRegressor
from autokeras.net_module import CnnModule, MlpModule
try:
    from autokeras.tabular.tabular_supervised import TabularClassifier, TabularRegressor
except OSError:
    print('You need to "brew install libomp".')



import csv
import itertools
import logging
import os
import pickle
import random
import string
import sys
import tempfile
import zipfile
from os import makedirs
from os.path import dirname
from os.path import exists
from sys import stdout

import imageio
import numpy as np
import requests
import torch
from scipy.ndimage import zoom

from autokeras.constant import Constant


class NoImprovementError(Exception):
    def __init__(self, message):
        self.message = message


def ensure_dir(directory):
    """Create directory if it does not exist."""
    if not os.path.exists(directory):
        os.makedirs(directory)


def ensure_file_dir(path):
    """Create path if it does not exist."""
    ensure_dir(os.path.dirname(path))


def has_file(path):
    """Check if the given path exists."""
    return os.path.exists(path)


def pickle_from_file(path):
    """Load the pickle file from the provided path and returns the object."""
    return pickle.load(open(path, 'rb'))


def pickle_to_file(obj, path):
    """Save the pickle file to the specified path."""
    pickle.dump(obj, open(path, 'wb'))


def get_device():
    """ If CUDA is available, use CUDA device, else use CPU device.
    Returns: string device name
    """
    return 'cuda' if torch.cuda.is_available() else 'cpu'


# # TODO cannot detect nvidia-smi in Windows normally. We need a fall back for windows
# def get_device():
#     """ If CUDA is available, use CUDA device, else use CPU device.
#     When choosing from CUDA devices, this function will choose the one with max memory available.
#     Returns: string device name.
#     """
#     # TODO: could use gputil in the future
#     device = 'cpu'
#     if torch.cuda.is_available():
#         try:
#             # smi_out=
#             #       Free                 : xxxxxx MiB
#             #       Free                 : xxxxxx MiB
#             #                      ....
#             smi_out = subprocess.check_output('nvidia-smi -q -d Memory | grep -A4 GPU|grep Free', shell=True)
#             if isinstance(smi_out, bytes):
#                 smi_out = smi_out.decode('utf-8')
#         except subprocess.SubprocessError:
#             warnings.warn('Cuda device successfully detected. However, nvidia-smi cannot be invoked')
#             return 'cpu'
#         visible_devices = os.getenv('CUDA_VISIBLE_DEVICES', '').split(',')
#         if len(visible_devices) == 1 and visible_devices[0] == '':
#             visible_devices = []
#         visible_devices = [int(x) for x in visible_devices]
#         memory_available = [int(x.split()[2]) for x in smi_out.splitlines()]
#         for cuda_index, _ in enumerate(memory_available):
#             if cuda_index not in visible_devices and visible_devices:
#                 memory_available[cuda_index] = 0
#         memory_available = list(filter(lambda a: a != 2, memory_available))
#         if memory_available:
#             if max(memory_available) != 0 and torch.cuda.device_count() == 1:
#                 device = 'cuda:' + str(memory_available.index(max(memory_available)))
#             elif max(memory_available) != 0 and torch.cuda.device_count() > 1:
#                 device = 'cuda:0'
#     return device


def temp_path_generator():
    sys_temp = tempfile.gettempdir()
    path = os.path.join(sys_temp, 'autokeras')
    return path


def rand_temp_folder_generator():
    """Create and return a temporary directory with the path name '/temp_dir_name/autokeras' (E:g:- /tmp/autokeras)."""
    chars = string.ascii_uppercase + string.digits
    size = 6
    random_suffix = ''.join(random.choice(chars) for _ in range(size))
    sys_temp = temp_path_generator()
    path = sys_temp + '_' + random_suffix
    ensure_dir(path)
    return path


def download_file(file_link, file_path):
    """Download the file specified in `file_link` and saves it in `file_path`."""
    if not os.path.exists(file_path):
        with open(file_path, "wb") as f:
            print("\nDownloading %s" % file_path)
            response = requests.get(file_link, stream=True)
            total_length = response.headers.get('content-length')

            if total_length is None:  # no content length header
                f.write(response.content)
            else:
                dl = 0
                total_length = int(total_length)
                for data in response.iter_content(chunk_size=4096):
                    dl += len(data)
                    f.write(data)
                    done = int(50 * dl / total_length)
                    sys.stdout.write("\r[%s%s]" % ('=' * done, ' ' * (50 - done)))
                    sys.stdout.flush()


def download_file_with_extract(file_link, file_path, extract_path):
    """Download the file specified in `file_link`, save to `file_path` and extract to the directory `extract_path`."""
    if not os.path.exists(extract_path):
        download_file(file_link, file_path)
        zip_ref = zipfile.ZipFile(file_path, 'r')
        print("extracting downloaded file...")
        zip_ref.extractall(extract_path)
        os.remove(file_path)
        print("extracted and removed downloaded zip file")
    print("file already extracted in the path %s" % extract_path)


def assert_search_space(search_space):
    grid = search_space
    value_list = []
    if Constant.LENGTH_DIM not in list(grid.keys()):
        print('No length dimension found in search Space. Using default values')
        grid[Constant.LENGTH_DIM] = Constant.DEFAULT_LENGTH_SEARCH
    elif not isinstance(grid[Constant.LENGTH_DIM][0], int):
        print('Converting String to integers. Next time please make sure to enter integer values for Length Dimension')
        grid[Constant.LENGTH_DIM] = list(map(int, grid[Constant.LENGTH_DIM]))

    if Constant.WIDTH_DIM not in list(grid.keys()):
        print('No width dimension found in search Space. Using default values')
        grid[Constant.WIDTH_DIM] = Constant.DEFAULT_WIDTH_SEARCH
    elif not isinstance(grid[Constant.WIDTH_DIM][0], int):
        print('Converting String to integers. Next time please make sure to enter integer values for Width Dimension')
        grid[Constant.WIDTH_DIM] = list(map(int, grid[Constant.WIDTH_DIM]))

    grid_key_list = list(grid.keys())
    grid_key_list.sort()
    for key in grid_key_list:
        value_list.append(grid[key])

    dimension = list(itertools.product(*value_list))
    # print(dimension)
    return grid, dimension


def verbose_print(new_father_id, new_graph, new_model_id):
    """Print information about the operation performed on father model to obtain current model and father's id."""
    cell_size = [24, 49]
    logging.info('New Model Id - ' + str(new_model_id))
    header = ['Father Model ID', 'Added Operation']
    line = '|'.join(str(x).center(cell_size[i]) for i, x in enumerate(header))
    logging.info('\n' + '+' + '-' * len(line) + '+')
    logging.info('|' + line + '|')
    logging.info('+' + '-' * len(line) + '+')
    for i in range(len(new_graph.operation_history)):
        if i == len(new_graph.operation_history) // 2:
            r = [str(new_father_id), ' '.join(str(item) for item in new_graph.operation_history[i])]
        else:
            r = [' ', ' '.join(str(item) for item in new_graph.operation_history[i])]
        line = '|'.join(str(x).center(cell_size[i]) for i, x in enumerate(r))
        logging.info('|' + line + '|')
    logging.info('+' + '-' * len(line) + '+')


def validate_xy(x_train, y_train):
    """Validate `x_train`'s type and the shape of `x_train`, `y_train`."""
    try:
        x_train = x_train.astype('float64')
    except ValueError:
        raise ValueError('x_train should only contain numerical data.')

    if len(x_train.shape) < 2:
        raise ValueError('x_train should at least has 2 dimensions.')

    if x_train.shape[0] != y_train.shape[0]:
        raise ValueError('x_train and y_train should have the same number of instances.')


def read_csv_file(csv_file_path):
    """Read the csv file and returns two separate list containing file names and their labels.

    Args:
        csv_file_path: Path to the CSV file.

    Returns:
        file_names: List containing files names.
        file_label: List containing their respective labels.
    """
    file_names = []
    file_labels = []
    with open(csv_file_path, 'r') as files_path:
        path_list = csv.DictReader(files_path)
        fieldnames = path_list.fieldnames
        for path in path_list:
            file_names.append(path[fieldnames[0]])
            file_labels.append(path[fieldnames[1]])
    return file_names, file_labels


def read_image(img_path):
    """Read the image contained in the provided path `image_path`."""
    img = imageio.imread(uri=img_path)
    return img


def compute_image_resize_params(data):
    """Compute median dimension of all images in data.

    It used to resize the images later. Number of channels do not change from the original data.

    Args:
        data: 1-D, 2-D or 3-D images. The Images are expected to have channel last configuration.

    Returns:
        median shape.
    """
    if data is None or len(data.shape) == 0:
        return []

    if len(data.shape) == len(data[0].shape) + 1 and np.prod(data[0].shape[:-1]) <= Constant.MAX_IMAGE_SIZE:
        return data[0].shape

    data_shapes = []
    for x in data:
        data_shapes.append(x.shape)

    median_shape = np.median(np.array(data_shapes), axis=0)
    median_size = np.prod(median_shape[:-1])

    if median_size > Constant.MAX_IMAGE_SIZE:
        reduction_factor = np.power(Constant.MAX_IMAGE_SIZE / median_size, 1 / (len(median_shape) - 1))
        median_shape[:-1] = median_shape[:-1] * reduction_factor

    return median_shape.astype(int)


def resize_image_data(data, resize_shape):
    """Resize images to given dimension.

    Args:
        data: 1-D, 2-D or 3-D images. The Images are expected to have channel last configuration.
        resize_shape: Image resize dimension.

    Returns:
        data: Reshaped data.
    """
    if data is None or len(resize_shape) == 0:
        return data

    if len(data.shape) > 1 and np.array_equal(data[0].shape, resize_shape):
        return data

    output_data = []
    for im in data:
        output_data.append(zoom(input=im, zoom=np.divide(resize_shape, im.shape)))

    return np.array(output_data)


def get_system():
    """Get the current system environment. If the current system is not supported, raise an exception.

    Returns:
         A string to represent the current OS name.
         "posix" stands for Linux, Mac or Solaris architecture.
         "nt" stands for Windows system.
    """
    if 'google.colab' in sys.modules:
        return Constant.SYS_GOOGLE_COLAB
    if os.name == 'posix':
        return Constant.SYS_LINUX
    if os.name == 'nt':
        return Constant.SYS_WINDOWS

    raise EnvironmentError('Unsupported environment')


def download_file_from_google_drive(file_id, dest_path, overwrite=False):
    """
    Downloads a shared file from google drive into a given folder.
    Optionally unzips it.

    Refact from:
    https://github.com/ndrplz/google-drive-downloader/blob/master/google_drive_downloader/google_drive_downloader.py

    Parameters
    ----------
    file_id: str
        the file identifier.
        You can obtain it from the sharable link.
    dest_path: str
        the destination where to save the downloaded file.
        Must be a path (for example: './downloaded_file.txt')
    overwrite: bool
        optional, if True forces re-download and overwrite.
    unzip: bool
        optional, if True unzips a file.
        If the file is not a zip file, ignores it.

    Returns
    -------
    None
    """

    destination_directory = dirname(dest_path)
    if len(destination_directory) > 0 and not exists(destination_directory):
        makedirs(destination_directory)

    if not exists(dest_path) or overwrite:

        session = requests.Session()

        print('Downloading file with Google ID {} into {}... '.format(file_id, dest_path), end='')
        stdout.flush()

        response = session.get(Constant.DOWNLOAD_URL, params={'id': file_id}, stream=True)

        token = get_confirm_token(response)
        if token:
            params = {'id': file_id, 'confirm': token}
            response = session.get(Constant.DOWNLOAD_URL, params=params, stream=True)

        save_response_content(response, dest_path)
        print('Done.')


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value
    return None


def save_response_content(response, destination):
    with open(destination, "wb") as f:
        for chunk in response.iter_content(Constant.CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
